/*
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * General Public License for more details.
 *
 */

/*
 * Copyright (c) 2000 VERITAS Software Corporation.
 *
 */
/****************************************************************************
 *  Header: remcom.c,v 1.34 91/03/09 12:29:49 glenne Exp $
 *
 *  Module name: remcom.c $
 *  Revision: 1.34 $
 *  Date: 91/03/09 12:29:49 $
 *  Contributor:     Lake Stevens Instrument Division$
 *
 *  Description:     low level support for gdb debugger. $
 *
 *  Considerations:  only works on target hardware $
 *
 *  Written by:	     Glenn Engel $
 *  Updated by:	     David Grothe <dave@gcom.com>
 *  ModuleState:     Experimental $
 *
 *  NOTES:	     See Below $
 *
 *  Modified for 386 by Jim Kingdon, Cygnus Support.
 *  Compatibility with 2.1.xx kernel by David Grothe <dave@gcom.com>
 *
 *  Changes to allow auto initilization.  All that is needed is that it
 *  be linked with the kernel and a break point (int 3) be executed.
 *  The header file <linux/kgdb.h> defines KGDB_BREAKPOINT to allow one to do
 *  this. It should also be possible, once the interrupt system is up, to
 *  call putDebugChar("+").  Once this is done, the remote debugger should
 *  get our attention by sending a ^C in a packet. George Anzinger
 *  <george@mvista.com>
 *  Integrated into 2.2.5 kernel by Tigran Aivazian <tigran@sco.com>
 *  Added thread support, support for multiple processors,
 *	support for ia-32(x86) hardware debugging.
 *	Amit S. Kale ( akale@veritas.com )
 *
 *
 *  To enable debugger support, two things need to happen.  One, a
 *  call to set_debug_traps() is necessary in order to allow any breakpoints
 *  or error conditions to be properly intercepted and reported to gdb.
 *  Two, a breakpoint needs to be generated to begin communication.  This
 *  is most easily accomplished by a call to kgdb_breakpoint().
 *  kgdb_breakpoint() simulates a breakpoint by executing an int 3.
 *
 *************
 *
 *    The following gdb commands are supported:
 *
 * command	    function				   Return value
 *
 *    g		    return the value of the CPU registers  hex data or ENN
 *    G		    set the value of the CPU registers	   OK or ENN
 *
 *    mAA..AA,LLLL  Read LLLL bytes at address AA..AA	   hex data or ENN
 *    MAA..AA,LLLL: Write LLLL bytes at address AA.AA	   OK or ENN
 *
 *    c		    Resume at current address		   SNN	 ( signal NN)
 *    cAA..AA	    Continue at address AA..AA		   SNN
 *
 *    s		    Step one instruction		   SNN
 *    sAA..AA	    Step one instruction from AA..AA	   SNN
 *
 *    k		    kill
 *
 *    ?		    What was the last sigval ?		   SNN	 (signal NN)
 *
 * All commands and responses are sent with a packet which includes a
 * checksum.  A packet consists of
 *
 * $<packet info>#<checksum>.
 *
 * where
 * <packet info> :: <characters representing the command or response>
 * <checksum>	 :: < two hex digits computed as modulo 256 sum of <packetinfo>>
 *
 * When a packet is received, it is first acknowledged with either '+' or '-'.
 * '+' indicates a successful transfer.	 '-' indicates a failed transfer.
 *
 * Example:
 *
 * Host:		  Reply:
 * $m0,10#2a		   +$00010203040506070809101112131415#42
 *
 ****************************************************************************/

#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/irq.h>
#include <linux/list.h>
#include <linux/console.h>
#include <linux/init.h>
#include <linux/kgdb.h>
#include <linux/fs.h>

#include <asm/uaccess.h>
#include <asm/semaphore.h>
#include <asm/vm86.h>
#include <asm/system.h>
#include <asm/ptrace.h>		/* for linux pt_regs struct */
#include <asm/kgdb_local.h>
#include <asm/atomic.h>
#include <asm/processor.h>
#include <asm/desc.h>
#include <asm/string.h>		/* for strcpy */

#ifdef CONFIG_SMP
#include <mach_ipi.h>
#endif

/************************************************************************
 *
 * external low-level support routines
 */
typedef void (*Function) (void);	/* pointer to a function */

/* Thread reference */
typedef unsigned char threadref[8];

int kgdb_enabled;
int kgdb_irq = -1;
EXPORT_SYMBOL(kgdb_irq);
int kgdb_pid_init_done;

extern void putDebugChar(int);	/* write a single character	 */
extern int getDebugChar(void);	/* read and return a single char */

/************************************************************************/
/* BUFMAX defines the maximum number of characters in inbound/outbound buffers*/
/* at least NUMREGBYTES*2 are needed for register packets */
/* Longer buffer is needed to list all threads */
#define BUFMAX 400

/*  debug >  0 prints ill-formed commands in valid packets & checksum errors */
int debug_regs = 0;		/* set to non-zero to print registers */

/* filled in by an external module */
char *gdb_module_offsets;

static const char hexchars[] = "0123456789abcdef";

/* Number of bytes of registers.  */
#define NUMREGBYTES 64
/*
 * Note that this register image is in a different order than
 * the register image that Linux produces at interrupt time.
 *
 * Linux's register image is defined by struct pt_regs in ptrace.h.
 * Just why GDB uses a different order is a historical mystery.
 */
enum regnames { _EAX,		/* 0 */
	_ECX,			/* 1 */
	_EDX,			/* 2 */
	_EBX,			/* 3 */
	_ESP,			/* 4 */
	_EBP,			/* 5 */
	_ESI,			/* 6 */
	_EDI,			/* 7 */
	_PC /* 8 also known as eip */ ,
	_PS /* 9 also known as eflags */ ,
	_CS,			/* 10 */
	_SS,			/* 11 */
	_DS,			/* 12 */
	_ES,			/* 13 */
	_FS,			/* 14 */
	_GS			/* 15 */
};

/***************************  ASSEMBLY CODE MACROS *************************/
/*
 * Put the error code here just in case the user cares.
 * Likewise, the vector number here (since GDB only gets the signal
 * number through the usual means, and that's not very specific).
 * The called_from is the return address so he can tell how we entered kgdb.
 * This will allow him to seperate out the various possible entries.
 */
#define REMOTE_DEBUG 0		/* set != to turn on printing (also available in info) */

#define PID_MAX PID_MAX_DEFAULT

#ifdef CONFIG_SMP
/*
 * By using the NMI code instead of a vector we just sneak thru the
 * word generator coming out with just what we want.  AND it does
 * not matter if clustered_apic_mode is set or not.
 */
static void smp_send_nmi_allbutself(void)
{
	send_IPI_allbutself(APIC_DM_NMI);
}
#else
static void smp_send_nmi_allbutself(void)
{
}
#endif

static struct kgdb_info {
	int used_malloc;
	void *called_from;
	long long entry_tsc;
	int errcode;
	int vector;
	int print_debug_info;
	int hold_on_sstep;
	struct {
		volatile struct task_struct *task;
		int pid;
		int hold;
		struct pt_regs *regs;
	} cpus_waiting[NR_CPUS];
} kgdb_info = {
	.hold_on_sstep = 1,
	.print_debug_info = REMOTE_DEBUG,
	.vector = -1,
};

#define used_m kgdb_info.used_malloc
/*
 * This is little area we set aside to contain the stack we
 * need to build to allow gdb to call functions.  We use one
 * per cpu to avoid locking issues.  We will do all this work
 * with interrupts off so that should take care of the protection
 * issues.
 */
#define LOOKASIDE_SIZE 200	/* should be more than enough */
#define MALLOC_MAX   200	/* Max malloc size */
struct {
	unsigned int esp;
	int array[LOOKASIDE_SIZE];
} fn_call_lookaside[NR_CPUS];

static int trap_cpu;
static unsigned int OLD_esp;
static int in_kgdb_console;

#define END_OF_LOOKASIDE  &fn_call_lookaside[trap_cpu].array[LOOKASIDE_SIZE]
#define IF_BIT 0x200
#define TF_BIT 0x100

static char gdbconbuf[BUFMAX];

static char *pack_hex_byte(char *pkt, int byte)
{
	*pkt++ = hexchars[(byte >> 4) & 0xf];
	*pkt++ = hexchars[(byte & 0xf)];
	return pkt;
}

static void putpacket(char *buffer)
{
	unsigned char checksum;
	int count;
	char ch;

	/*  $<packet info>#<checksum>. */
	do {
		if (kgdb_info.print_debug_info)
			printk("T:%s\n", buffer);
		putDebugChar('$');
		checksum = 0;
		count = 0;

		while ((ch = buffer[count])) {
			putDebugChar(ch);
			checksum += ch;
			count += 1;
		}

		putDebugChar('#');
		putDebugChar(hexchars[checksum >> 4]);
		putDebugChar(hexchars[checksum % 16]);

	} while ((getDebugChar() & 0x7f) != '+');

}

static void kgdb_gdb_message(const char *s, unsigned count)
{
	int i;
	int wcount;
	char *bufptr;

	/*
	 * This takes care of NMI while spining out chars to gdb
	 */
	in_kgdb_console = 1;
	gdbconbuf[0] = 'O';
	bufptr = gdbconbuf + 1;
	while (count > 0) {
		if ((count << 1) > (BUFMAX - 2))
			wcount = (BUFMAX - 2) >> 1;
		else
			wcount = count;
		count -= wcount;
		for (i = 0; i < wcount; i++)
			bufptr = pack_hex_byte(bufptr, s[i]);
		*bufptr = '\0';
		s += wcount;
		putpacket(gdbconbuf);
	}
	in_kgdb_console = 0;
}

static void to_gdb(const char *s)
{
	int count = 0;

	while (s[count] && (count++ < BUFMAX))
		;
	kgdb_gdb_message(s, count);
}

/*
 * Gdb calls functions by pushing agruments, including a return address
 * on the stack and the adjusting EIP to point to the function.	 The
 * whole assumption in GDB is that we are on a different stack than the
 * one the "user" i.e. code that hit the break point, is on.  This, of
 * course is not true in the kernel.  Thus various dodges are needed to
 * do the call without directly messing with EIP (which we can not change
 * as it is just a location and not a register.	 To adjust it would then
 * require that we move every thing below EIP up or down as needed.  This
 * will not work as we may well have stack relative pointer on the stack
 * (such as the pointer to regs, for example).

 * So here is what we do:
 * We detect gdb attempting to store into the stack area and instead, store
 * into the fn_call_lookaside.array at the same relative location as if it
 * were the area ESP pointed at.  We also trap ESP modifications
 * and uses these to adjust fn_call_lookaside.esp.  On entry
 * fn_call_lookaside.esp will be set to point at the last entry in
 * fn_call_lookaside.array.  This allows us to check if it has changed, and
 * if so, on exit, we add the registers we will use to do the move and a
 * trap/ interrupt return exit sequence.  We then adjust the eflags in the
 * regs array (remember we now have a copy in the fn_call_lookaside.array) to
 * kill the interrupt bit, AND we change EIP to point at our set up stub.
 * As part of the register set up we preset the registers to point at the
 * begining and end of the fn_call_lookaside.array, so all the stub needs to
 * do is move words from the array to the stack until ESP= the desired value
 * then do the rti.  This will then transfer to the desired function with
 * all the correct registers.  Nifty huh?
 */
extern asmlinkage void fn_call_stub(void);
extern asmlinkage void fn_rtn_stub(void);

__asm__("fn_rtn_stub:\n\t"
	"movl %eax,%esp\n\t"
	"fn_call_stub:\n\t"
	"1:\n\t"
	"addl $-4,%ebx\n\t"
	"movl (%ebx), %eax\n\t"
	"pushl %eax\n\t"
	"cmpl %esp,%ecx\n\t"
	"jne  1b\n\t"
	"popl %eax\n\t"
	"popl %ebx\n\t"
	"popl %ecx\n\t"
	"iret \n\t");

spinlock_t kgdb_spinlock = SPIN_LOCK_UNLOCKED;

static spinlock_t waitlocks[NR_CPUS] =
		{[0 ... NR_CPUS - 1] = SPIN_LOCK_UNLOCKED };
volatile struct pt_regs *in_kgdb_here_log[NR_CPUS];

/*
 * The following array has the thread pointer of each of the "other"
 * cpus.  We make it global so it can be seen by gdb.
 */
volatile int in_kgdb_entry_log[NR_CPUS];

#ifdef CONFIG_SMP

/* waiters on our spinlock plus us */
static atomic_t spinlock_waiters = ATOMIC_INIT(1);
static int spinlock_count = 0;
static int spinlock_cpu = 0;

/*
 * Note we use nested spin locks to account for the case where a break
 * point is encountered when calling a function by user direction from
 * kgdb. Also there is the memory exception recursion to account for.
 * Well, yes, but this lets other cpus thru too.  Lets add a
 * cpu id to the lock.
 */
#define KGDB_SPIN_LOCK(x) if (spinlock_count == 0 ||			\
			      spinlock_cpu != smp_processor_id()){	\
				      atomic_inc(&spinlock_waiters);	\
				      while (! spin_trylock(x)) {	\
					    in_kgdb(regs);		\
				      }					\
				      atomic_dec(&spinlock_waiters);	\
				      spinlock_count = 1;		\
				      spinlock_cpu = smp_processor_id();\
			  } else {					\
				      spinlock_count++;			\
			  }
#define KGDB_SPIN_UNLOCK(x) if (--spinlock_count == 0) spin_unlock(x)
#else
static unsigned kgdb_depth = 0;
#define KGDB_SPIN_LOCK(x) do { kgdb_depth--; } while (0)
#define KGDB_SPIN_UNLOCK(x) do { kgdb_depth++; } while (0)
#define spinlock_count 0
#endif

static int hex(char ch)
{
	if ((ch >= 'a') && (ch <= 'f'))
		return (ch - 'a' + 10);
	if ((ch >= '0') && (ch <= '9'))
		return (ch - '0');
	if ((ch >= 'A') && (ch <= 'F'))
		return (ch - 'A' + 10);
	return (-1);
}

/* scan for the sequence $<data>#<checksum>	*/
static void getpacket(char *buffer)
{
	unsigned char checksum;
	unsigned char xmitcsum;
	int i;
	int count;
	char ch;

	do {
		/*
		 * wait around for the start character, ignore all other
		 * characters
		 */
		while ((ch = (getDebugChar() & 0x7f)) != '$') ;
		checksum = 0;
		xmitcsum = -1;

		count = 0;

		/* now, read until a # or end of buffer is found */
		while (count < BUFMAX) {
			ch = getDebugChar() & 0x7f;
			if (ch == '#')
				break;
			checksum = checksum + ch;
			buffer[count] = ch;
			count = count + 1;
		}
		buffer[count] = 0;

		if (ch == '#') {
			xmitcsum = hex(getDebugChar() & 0x7f) << 4;
			xmitcsum += hex(getDebugChar() & 0x7f);
			if ((kgdb_info.print_debug_info) &&
					(checksum != xmitcsum)) {
				printk("bad checksum.	My count = 0x%x, "
						"sent=0x%x. buf=%s\n",
					checksum, xmitcsum, buffer);
			}

			if (checksum != xmitcsum)
				putDebugChar('-');	/* failed checksum */
			else {
				putDebugChar('+'); /* successful transfer */
				/*
				 * if a sequence char is present, reply the
				 * sequence ID
				 */
				if (buffer[2] == ':') {
					putDebugChar(buffer[0]);
					putDebugChar(buffer[1]);
					/* remove sequence chars from buffer */
					count = strlen(buffer);
					for (i = 3; i <= count; i++)
						buffer[i - 3] = buffer[i];
				}
			}
		}
	} while (checksum != xmitcsum);

	if (kgdb_info.print_debug_info)
		printk("R:%s\n", buffer);
}

/* send the packet in buffer.  */

static char remcomInBuffer[BUFMAX];
static char remcomOutBuffer[BUFMAX];
static short error;

static void debug_error(char *format, char *parm)
{
	if (kgdb_info.print_debug_info)
		printk(format, parm);
}

static void print_regs(struct pt_regs *regs)
{
	printk("EAX=%08lx ", regs->eax);
	printk("EBX=%08lx ", regs->ebx);
	printk("ECX=%08lx ", regs->ecx);
	printk("EDX=%08lx ", regs->edx);
	printk("\n");
	printk("ESI=%08lx ", regs->esi);
	printk("EDI=%08lx ", regs->edi);
	printk("EBP=%08lx ", regs->ebp);
	printk("ESP=%08lx ", (long) &regs->esp);
	printk("\n");
	printk(" DS=%08x ", regs->xds);
	printk(" ES=%08x ", regs->xes);
	printk(" SS=%08x ", __KERNEL_DS);
	printk(" FL=%08lx ", regs->eflags);
	printk("\n");
	printk(" CS=%08x ", regs->xcs);
	printk(" IP=%08lx ", regs->eip);
	printk("\n");

}

static void regs_to_gdb_regs(int *gdb_regs, struct pt_regs *regs)
{
	gdb_regs[_EAX] = regs->eax;
	gdb_regs[_EBX] = regs->ebx;
	gdb_regs[_ECX] = regs->ecx;
	gdb_regs[_EDX] = regs->edx;
	gdb_regs[_ESI] = regs->esi;
	gdb_regs[_EDI] = regs->edi;
	gdb_regs[_EBP] = regs->ebp;
	gdb_regs[_DS] = regs->xds;
	gdb_regs[_ES] = regs->xes;
	gdb_regs[_PS] = regs->eflags;
	gdb_regs[_CS] = regs->xcs;
	gdb_regs[_PC] = regs->eip;
	/* Note, as we are a debugging the kernel, we will always
	 * trap in kernel code, this means no priviledge change,
	 * and so the pt_regs structure is not completely valid.  In a non
	 * privilege change trap, only EFLAGS, CS and EIP are put on the stack,
	 * SS and ESP are not stacked, this means that the last 2 elements of
	 * pt_regs is not valid (they would normally refer to the user stack)
	 * also, using regs+1 is no good because you end up will a value that is
	 * 2 longs (8) too high.  This used to cause stepping over functions
	 * to fail, so my fix is to use the address of regs->esp, which
	 * should point at the end of the stack frame.	Note I have ignored
	 * completely exceptions that cause an error code to be stacked, such
	 * as double fault.  Stuart Hughes, Zentropix.
	 * original code: gdb_regs[_ESP] =  (int) (regs + 1) ;

	 * this is now done on entry and moved to OLD_esp (as well as
	 * fn_call_lookaside[trap_cpu].esp).
	 */
	gdb_regs[_ESP] = fn_call_lookaside[trap_cpu].esp;
	gdb_regs[_SS] = __KERNEL_DS;
	gdb_regs[_FS] = 0xFFFF;
	gdb_regs[_GS] = 0xFFFF;
}

static void gdb_regs_to_regs(int *gdb_regs, struct pt_regs *regs)
{
	regs->eax = gdb_regs[_EAX];
	regs->ebx = gdb_regs[_EBX];
	regs->ecx = gdb_regs[_ECX];
	regs->edx = gdb_regs[_EDX];
	regs->esi = gdb_regs[_ESI];
	regs->edi = gdb_regs[_EDI];
	regs->ebp = gdb_regs[_EBP];
	regs->xds = gdb_regs[_DS];
	regs->xes = gdb_regs[_ES];
	regs->eflags = gdb_regs[_PS];
	regs->xcs = gdb_regs[_CS];
	regs->eip = gdb_regs[_PC];
	fn_call_lookaside[trap_cpu].esp = gdb_regs[_ESP]; /* keep value */
#if 0	/* can't change these */
	regs->esp = gdb_regs[_ESP];
	regs->xss = gdb_regs[_SS];
	regs->fs = gdb_regs[_FS];
	regs->gs = gdb_regs[_GS];
#endif

}

static int thread_list = 0;

void get_gdb_regs(struct task_struct *p, struct pt_regs *regs, int *gdb_regs)
{
	unsigned long stack_page;
	int i;

	if (!p || p == current) {
		regs_to_gdb_regs(gdb_regs, regs);
		return;
	}
#ifdef CONFIG_SMP
	for_each_online_cpu(i) {
		if (p != kgdb_info.cpus_waiting[i].task)
			continue;
		regs_to_gdb_regs(gdb_regs, kgdb_info.cpus_waiting[i].regs);
		gdb_regs[_ESP] = (int)&kgdb_info.cpus_waiting[i].regs->esp;
		return;
	}
#endif
	memset(gdb_regs, 0, NUMREGBYTES);
	gdb_regs[_ESP] = p->thread.esp;
	gdb_regs[_PC] = p->thread.eip;
	gdb_regs[_EBP] = *(int *) gdb_regs[_ESP];
	gdb_regs[_EDI] = *(int *) (gdb_regs[_ESP] + 4);
	gdb_regs[_ESI] = *(int *) (gdb_regs[_ESP] + 8);

	/*
	 * This code is to give a more informative notion of where a process
	 * is waiting.	It is used only when the user asks for a thread info
	 * list.  If he then switches to the thread, s/he will find the task
	 * is in schedule, but a back trace should show the same info we come
	 * up with.  This code was shamelessly purloined from process.c.  It was
	 * then enhanced to provide more registers than simply the program
	 * counter.
	 */
	if (!thread_list)
		return;

	if (p->state == TASK_RUNNING)
		return;
	stack_page = (unsigned long)p->thread_info;
	if (gdb_regs[_ESP] < stack_page ||
			gdb_regs[_ESP] > THREAD_SIZE-sizeof(long)+stack_page)
		return;
	/* include/asm-i386/system.h:switch_to() pushes ebp last. */
	i = 0;
	do {
		if (gdb_regs[_EBP] < stack_page ||
		    gdb_regs[_EBP] > THREAD_SIZE - 2*sizeof(long) + stack_page)
			return;
		gdb_regs[_PC] = *(unsigned long *)(gdb_regs[_EBP] + 4);
		gdb_regs[_ESP] = gdb_regs[_EBP] + 8;
		gdb_regs[_EBP] = *(unsigned long *)gdb_regs[_EBP];
		if (!in_sched_functions(gdb_regs[_PC]))
			return;
	} while (i++ < 16);
	return;
}

/* Indicate to caller of mem2hex or hex2mem that there has been an error  */
static volatile int mem_err = 0;
static volatile int mem_err_expected = 0;
static volatile int mem_err_cnt = 0;
static int garbage_loc = -1;

static int get_char(char *addr)
{
	return *addr;
}

static void set_char(char *addr, int val, int may_fault)
{
	/*
	 * This code traps references to the area mapped to the kernel
	 * stack as given by the regs and, instead, stores to the
	 * fn_call_lookaside[cpu].array
	 */
	if (may_fault &&
	    (unsigned int) addr < OLD_esp &&
	    ((unsigned int) addr > (OLD_esp - (unsigned int) LOOKASIDE_SIZE))) {
		addr = (char *) END_OF_LOOKASIDE - ((char *) OLD_esp - addr);
	}
	*addr = val;
}

/*
 * convert the memory pointed to by mem into hex, placing result in buf
 * return a pointer to the last char put in buf (null)
 * If MAY_FAULT is non-zero, then we should set mem_err in response to
 *  a fault; if zero treat a fault like any other fault in the stub.
 */
static char *mem2hex(char *mem, char *buf, int count, int may_fault)
{
	int i;
	unsigned char ch;

	if (may_fault) {
		mem_err_expected = 1;
		mem_err = 0;
	}
	for (i = 0; i < count; i++) {
		ch = get_char(mem++);
		if (may_fault && mem_err) {
			if (kgdb_info.print_debug_info)
				printk("Mem fault fetching from addr %lx\n",
				       (long) (mem - 1));
			*buf = 0;	/* truncate buffer */
			return (buf);
		}
		*buf++ = hexchars[ch >> 4];
		*buf++ = hexchars[ch % 16];
	}
	*buf = 0;
	if (may_fault)
		mem_err_expected = 0;
	return (buf);
}

/* convert the hex array pointed to by buf into binary to be placed in mem */
/* return a pointer to the character AFTER the last byte written */
/* NOTE: We use the may fault flag to also indicate if the write is to
 * the registers (0) or "other" memory (!=0)
 */
static char *hex2mem(char *buf, char *mem, int count, int may_fault)
{
	int i;
	unsigned char ch;

	if (may_fault) {
		mem_err_expected = 1;
		mem_err = 0;
	}
	for (i = 0; i < count; i++) {
		ch = hex(*buf++) << 4;
		ch = ch + hex(*buf++);
		set_char(mem++, ch, may_fault);

		if (may_fault && mem_err) {
			if (kgdb_info.print_debug_info)
				printk("Mem fault storing to addr %lx\n",
				       (long) (mem - 1));
			return (mem);
		}
	}
	if (may_fault)
		mem_err_expected = 0;
	return (mem);
}

static int hexToInt(char **ptr, int *intValue)
{
	int numChars = 0;
	int hexValue;

	*intValue = 0;

	while (**ptr) {
		hexValue = hex(**ptr);
		if (hexValue >= 0) {
			*intValue = (*intValue << 4) | hexValue;
			numChars++;
		} else
			break;

		(*ptr)++;
	}

	return (numChars);
}

static char *pack_threadid(char *pkt, threadref *id)
{
	char *limit;
	unsigned char *altid;

	altid = (unsigned char *)id;
	limit = pkt + 16;
	while (pkt < limit)
		pkt = pack_hex_byte(pkt, *altid++);
	return pkt;
}

static void int_to_threadref(threadref *id, int value)
{
	unsigned char *scan;
	int i;

	scan = (unsigned char *)id;
	i = 4;
	while (i--)
		*scan++ = 0;
	*scan++ = (value >> 24) & 0xff;
	*scan++ = (value >> 16) & 0xff;
	*scan++ = (value >> 8) & 0xff;
	*scan++ = (value & 0xff);
}

static int int_to_hex_v(unsigned char *id, int value)
{
	unsigned char *start = id;
	int shift;
	int ch;

	for (shift = 28; shift >= 0; shift -= 4) {
		if ((ch = (value >> shift) & 0xf) || (id != start)) {
			*id = hexchars[ch];
			id++;
		}
	}
	if (id == start)
		*id++ = '0';
	return id - start;
}

static int cmp_str(char *s1, char *s2, int count)
{
	while (count--) {
		if (*s1++ != *s2++)
			return 0;
	}
	return 1;
}

struct task_struct *getthread(int pid)
{
	struct task_struct *thread;

	if (pid >= PID_MAX && pid <= (PID_MAX + NR_CPUS)) {
		return kgdb_get_idle(pid - PID_MAX);
	} else {
		/*
		 * find_task_by_pid is relatively safe all the time
		 * Other pid functions require lock downs which imply
		 * that we may be interrupting them (as we get here
		 * in the middle of most any lock down).
		 * Still we don't want to call until the table exists!
		 */
		if (kgdb_pid_init_done) {
			thread = find_task_by_pid(pid);
			if (thread)
				return thread;
		}
	}
	return NULL;
}

static struct hw_breakpoint {
	unsigned enabled;
	unsigned type;
	unsigned len;
	unsigned addr;
} breakinfo[4];

static unsigned hw_breakpoint_status;

static void correct_hw_break(void)
{
	int breakno;
	int correctit;
	int breakbit;
	unsigned dr7;

	asm volatile ("movl %%db7, %0\n":"=r" (dr7) : );
	do {
		unsigned addr0, addr1, addr2, addr3;
		asm volatile ("movl %%db0, %0\n"
			      "movl %%db1, %1\n"
			      "movl %%db2, %2\n"
			      "movl %%db3, %3\n"
			      :"=r" (addr0), "=r"(addr1),
			      "=r"(addr2), "=r"(addr3)
			      :);
	} while (0);
	correctit = 0;
	for (breakno = 0; breakno < 3; breakno++) {
		breakbit = 2 << (breakno << 1);
		if (!(dr7 & breakbit) && breakinfo[breakno].enabled) {
			correctit = 1;
			dr7 |= breakbit;
			dr7 &= ~(0xf0000 << (breakno << 2));
			dr7 |= (((breakinfo[breakno].len << 2) |
				 breakinfo[breakno].type) << 16) <<
			    (breakno << 2);
			switch (breakno) {
			case 0:
				asm volatile ("movl %0, %%dr0\n"::"r"
					      (breakinfo[breakno].addr));
				break;

			case 1:
				asm volatile ("movl %0, %%dr1\n"::"r"
					      (breakinfo[breakno].addr));
				break;

			case 2:
				asm volatile ("movl %0, %%dr2\n"::"r"
					      (breakinfo[breakno].addr));
				break;

			case 3:
				asm volatile ("movl %0, %%dr3\n"::"r"
					      (breakinfo[breakno].addr));
				break;
			}
		} else if ((dr7 & breakbit) && !breakinfo[breakno].enabled) {
			correctit = 1;
			dr7 &= ~breakbit;
			dr7 &= ~(0xf0000 << (breakno << 2));
		}
	}
	if (correctit)
		asm volatile ("movl %0, %%db7\n"::"r" (dr7));
}

static int remove_hw_break(unsigned breakno)
{
	if (!breakinfo[breakno].enabled)
		return -1;
	breakinfo[breakno].enabled = 0;
	return 0;
}

static int set_hw_break(unsigned breakno, unsigned type, unsigned len,
			unsigned addr)
{
	if (breakinfo[breakno].enabled)
		return -1;
	breakinfo[breakno].enabled = 1;
	breakinfo[breakno].type = type;
	breakinfo[breakno].len = len;
	breakinfo[breakno].addr = addr;
	return 0;
}

#ifdef CONFIG_SMP

int in_kgdb(struct pt_regs *regs)
{
	unsigned long flags;
	int cpu = smp_processor_id();

	if (!spin_is_locked(&kgdb_spinlock)) {
		if (in_kgdb_here_log[cpu] ||	/* we are holding this cpu */
		    in_kgdb_console) {	/* or we are doing slow i/o */
			return 1;
		}
		return 0;
	}

	/* As I see it the only reason not to let all cpus spin on
	 * the same spin_lock is to allow selected ones to proceed.
	 * This would be a good thing, so we leave it this way.
	 * Maybe someday....  Done !

	 * in_kgdb() is called from an NMI so we don't pretend
	 * to have any resources, like printk() for example.
	 */

	local_irq_save(flags);	/* only local here, to avoid hanging */
	/*
	 * log arival of this cpu
	 * The NMI keeps on ticking.  Protect against recurring more
	 * than once, and ignor the cpu that has the kgdb lock
	 */
	in_kgdb_entry_log[cpu]++;
	in_kgdb_here_log[cpu] = regs;
	if (cpu == spinlock_cpu || kgdb_info.cpus_waiting[cpu].task) {
		goto exit_in_kgdb;
	}
	/*
	 * For protection of the initilization of the spin locks by kgdb
	 * it locks the kgdb spinlock before it gets the wait locks set
	 * up.	We wait here for the wait lock to be taken.  If the
	 * kgdb lock goes away first??	Well, it could be a slow exit
	 * sequence where the wait lock is removed prior to the kgdb lock
	 * so if kgdb gets unlocked, we just exit.
	 */
	while (spin_is_locked(&kgdb_spinlock) &&
	       !spin_is_locked(waitlocks + cpu)) ;
	if (!spin_is_locked(&kgdb_spinlock)) {
		goto exit_in_kgdb;
	}
	kgdb_info.cpus_waiting[cpu].task = current;
	kgdb_info.cpus_waiting[cpu].pid = (current->pid) ? : (PID_MAX + cpu);
	kgdb_info.cpus_waiting[cpu].regs = regs;

	spin_unlock_wait(waitlocks + cpu);
	/*
	 * log departure of this cpu
	 */
	kgdb_info.cpus_waiting[cpu].task = 0;
	kgdb_info.cpus_waiting[cpu].pid = 0;
	kgdb_info.cpus_waiting[cpu].regs = 0;
	correct_hw_break();
exit_in_kgdb:
	in_kgdb_here_log[cpu] = 0;
	local_irq_restore(flags);
	return 1;
	/*
	   spin_unlock(continuelocks + smp_processor_id());
	 */
}

#else

int in_kgdb(struct pt_regs *regs)
{
	return kgdb_depth;
}
#endif

static void printexceptioninfo(int exceptionNo, int errorcode, char *buffer)
{
	unsigned dr6;
	int i;

	switch (exceptionNo) {
	case 1:		/* debug exception */
		break;
	case 3:		/* breakpoint */
		sprintf(buffer, "Software breakpoint");
		return;
	default:
		sprintf(buffer, "Details not available");
		return;
	}
	asm volatile ("movl %%db6, %0\n":"=r" (dr6) :);
	if (dr6 & 0x4000) {
		sprintf(buffer, "Single step");
		return;
	}
	for (i = 0; i < 4; ++i) {
		if (dr6 & (1 << i)) {
			sprintf(buffer, "Hardware breakpoint %d", i);
			return;
		}
	}
	sprintf(buffer, "Unknown trap");
	return;
}

/*
 * This function does all command procesing for interfacing to gdb.
 *
 * NOTE:  The INT nn instruction leaves the state of the interrupt
 *	  enable flag UNCHANGED.  That means that when this routine
 *	  is entered via a breakpoint (INT 3) instruction from code
 *	  that has interrupts enabled, then interrupts will STILL BE
 *	  enabled when this routine is entered.	 The first thing that
 *	  we do here is disable interrupts so as to prevent recursive
 *	  entries and bothersome serial interrupts while we are
 *	  trying to run the serial port in polled mode.
 *
 * For kernel version 2.1.xx the kgdb_cli() actually gets a spin lock so
 * it is always necessary to do a restore_flags before returning
 * so as to let go of that lock.
 */
void kgdb_handle_exception(int exceptionVector, int signo, int err_code,
			struct pt_regs *regs)
{
	struct task_struct *usethread = NULL;
	struct task_struct *thread_list_start = 0, *thread = NULL;
	int addr, length;
	int breakno, breaktype;
	char *ptr;
	int newPC;
	threadref thref;
	int threadid;
	int thread_min = PID_MAX + NR_CPUS;
	int nothreads;
	unsigned long flags;
	int gdb_regs[NUMREGBYTES / 4];
	int dr6;
	int entry_state = 0;	/* 0, ok, 1, no nmi, 2 sync failed */
	static long no_idt[2];
#define NO_NMI 1
#define NO_SYNC 2
#define NUMREGS NUMREGBYTES/4
	/*
	 * If the entry is not from the kernel then return to the Linux
	 * trap handler and let it process the interrupt normally.
	 */
	if ((regs->eflags & VM_MASK) || (3 & regs->xcs)) {
		printk("ignoring non-kernel exception\n");
		print_regs(regs);
		return;
	}

	local_irq_save(flags);

	/* Get kgdb spinlock */

	KGDB_SPIN_LOCK(&kgdb_spinlock);
	rdtscll(kgdb_info.entry_tsc);
	/*
	 * We depend on this spinlock and the NMI watch dog to control the
	 * other cpus.	They will arrive at "in_kgdb()" as a result of the
	 * NMI and will wait there for the following spin locks to be
	 * released.
	 */
	if (spinlock_count == 1) {
		int time = 0, end_time, dum = 0;
		int i;
		int cpu_logged_in[NR_CPUS] = { 0 };

		if (kgdb_info.print_debug_info)
			printk("kgdb : cpu %d entry, syncing others\n",
			       smp_processor_id());
		for_each_online_cpu(i) {
			/*
			 * Use trylock as we may already hold the lock if
			 * we are holding the cpu.  Net result is all
			 * locked.
			 */
			spin_trylock(&waitlocks[i]);
		}
		for_each_online_cpu(i)
			cpu_logged_in[i] = 0;
		/*
		 * Wait for their arrival.  We know the watch dog is active if
		 * in_kgdb() has ever been called, as it is always called on a
		 * watchdog tick.
		 */
		rdtsc(dum, time);
		end_time = time + 2;	/* Note: we use the High order bits! */
		i = 1;
		if (num_online_cpus() > 1) {
			int me_in_kgdb = in_kgdb_entry_log[smp_processor_id()];
			smp_send_nmi_allbutself();
			while (i < num_online_cpus() && time != end_time) {
				int j;

				for_each_online_cpu(j) {
					if (kgdb_info.cpus_waiting[j].task &&
					    !cpu_logged_in[j]) {
						i++;
						cpu_logged_in[j] = 1;
						if (kgdb_info.print_debug_info) {
							printk
							    ("kgdb : cpu %d arrived at kgdb\n",
							     j);
						}
						break;
					} else if (!kgdb_info.cpus_waiting[j].task &&
						   !cpu_online(j)) {
						kgdb_info.cpus_waiting[j].task =
							(struct task_struct *)
								0xbad1fbad;
						cpu_logged_in[j] = 1;
						kgdb_info.cpus_waiting[j].hold = 1;
						break;
					}
					if (!kgdb_info.cpus_waiting[j].task &&
					    in_kgdb_here_log[j]) {

						int wait = 100000;
						while (wait--) ;
						if (!kgdb_info.cpus_waiting[j].task &&
						    in_kgdb_here_log[j]) {
							printk
							    ("kgdb : cpu %d stall"
							     " in in_kgdb\n",
							     j);
							i++;
							cpu_logged_in[j] = 1;
							kgdb_info.cpus_waiting[j].task =
							    (struct task_struct
							     *) 1;
						}
					}
				}

				if (in_kgdb_entry_log[smp_processor_id()] >
				    (me_in_kgdb + 10)) {
					break;
				}

				rdtsc(dum, time);
			}
			if (i < num_online_cpus()) {
				printk
				    ("kgdb : time out, proceeding without sync\n");
				entry_state = NO_SYNC;
			}
		} else {
			if (kgdb_info.print_debug_info) {
				printk
				    ("kgdb : %d cpus, but watchdog not active\n"
				     "proceeding without locking down other cpus\n",
				     num_online_cpus());
				entry_state = NO_NMI;
			}
		}
	}

	if (kgdb_info.print_debug_info) {
		unsigned long *lp = (unsigned long *) &regs;

		printk("handle_exception(exceptionVector=%d, "
		       "signo=%d, err_code=%d, regs=%p)\n",
		       exceptionVector, signo, err_code, regs);
		if (debug_regs) {
			print_regs(regs);
			printk("Stk: %8lx %8lx %8lx %8lx"
			       "  %8lx %8lx %8lx %8lx\n",
			       lp[0], lp[1], lp[2], lp[3],
			       lp[4], lp[5], lp[6], lp[7]);
			printk("     %8lx %8lx %8lx %8lx"
			       "  %8lx %8lx %8lx %8lx\n",
			       lp[8], lp[9], lp[10], lp[11],
			       lp[12], lp[13], lp[14], lp[15]);
			printk("     %8lx %8lx %8lx %8lx  "
			       "%8lx %8lx %8lx %8lx\n",
			       lp[16], lp[17], lp[18], lp[19],
			       lp[20], lp[21], lp[22], lp[23]);
			printk("     %8lx %8lx %8lx %8lx  "
			       "%8lx %8lx %8lx %8lx\n",
			       lp[24], lp[25], lp[26], lp[27],
			       lp[28], lp[29], lp[30], lp[31]);
		}
	}

	/* Disable hardware debugging while we are in kgdb */
	/* Get the debug register status register */
      __asm__("movl %0,%%db7" :	/* no output */ :"r"(0));

	asm volatile ("movl %%db6, %0\n" :"=r" (hw_breakpoint_status) :);

	switch (exceptionVector) {
	case 0:		/* divide error */
	case 1:		/* debug exception */
	case 2:		/* NMI */
	case 3:		/* breakpoint */
	case 4:		/* overflow */
	case 5:		/* bounds check */
	case 6:		/* invalid opcode */
	case 7:		/* device not available */
	case 8:		/* double fault (errcode) */
	case 10:		/* invalid TSS (errcode) */
	case 12:		/* stack fault (errcode) */
	case 16:		/* floating point error */
	case 17:		/* alignment check (errcode) */
	default:		/* any undocumented */
		break;
	case 11:		/* segment not present (errcode) */
	case 13:		/* general protection (errcode) */
	case 14:		/* page fault (special errcode) */
	case 19:		/* cache flush denied */
		if (mem_err_expected) {
			/*
			 * This fault occured because of the
			 * get_char or set_char routines.  These
			 * two routines use either eax of edx to
			 * indirectly reference the location in
			 * memory that they are working with.
			 * For a page fault, when we return the
			 * instruction will be retried, so we
			 * have to make sure that these
			 * registers point to valid memory.
			 */
			mem_err = 1;	/* set mem error flag */
			mem_err_expected = 0;
			mem_err_cnt++;	/* helps in debugging */
			/* make valid address */
			regs->eax = (long)&garbage_loc;
			/* make valid address */
			regs->edx = (long)&garbage_loc;
			if (kgdb_info.print_debug_info)
				printk("Return after memory error: "
				       "mem_err_cnt=%d\n", mem_err_cnt);
			if (debug_regs)
				print_regs(regs);
			goto exit_kgdb;
		}
		break;
	}
	if (kgdb_info.print_debug_info)
		printk("kgdb : entered kgdb on cpu %d\n", smp_processor_id());

	kgdb_info.vector = exceptionVector;
	kgdb_info.errcode = err_code;
	kgdb_info.called_from = __builtin_return_address(0);
#ifdef CONFIG_SMP
	/*
	 * OK, we can now communicate, lets tell gdb about the sync.
	 * but only if we had a problem.
	 */
	switch (entry_state) {
	case NO_NMI:
		to_gdb("NMI not active, other cpus not stopped\n");
		break;
	case NO_SYNC:
		to_gdb("Some cpus not stopped, see 'kgdb_info' for details\n");
	default:;
	}

#endif
	/*
	 * Set up the gdb function call area.
	 */
	trap_cpu = smp_processor_id();
	OLD_esp = fn_call_lookaside[trap_cpu].esp = (int)(&regs->esp);

once_again:
	    /* reply to host that an exception has occurred */
	remcomOutBuffer[0] = 'S';
	remcomOutBuffer[1] = hexchars[signo >> 4];
	remcomOutBuffer[2] = hexchars[signo % 16];
	remcomOutBuffer[3] = 0;

	putpacket(remcomOutBuffer);

	while (1) {
		error = 0;
		remcomOutBuffer[0] = 0;
		getpacket(remcomInBuffer);
		switch (remcomInBuffer[0]) {
		case '?':
			remcomOutBuffer[0] = 'S';
			remcomOutBuffer[1] = hexchars[signo >> 4];
			remcomOutBuffer[2] = hexchars[signo % 16];
			remcomOutBuffer[3] = 0;
			break;
		case 'd':
			kgdb_info.print_debug_info = !(kgdb_info.print_debug_info);	/* toggle debug flag */
			printk("Remote debug %s\n",
			       kgdb_info.print_debug_info ? "on" : "off");
			break;
		case 'g':	/* return the value of the CPU registers */
			get_gdb_regs(usethread, regs, gdb_regs);
			mem2hex((char *)gdb_regs,
				remcomOutBuffer, NUMREGBYTES, 0);
			break;
		/* set the value of the CPU registers - return OK */
		case 'G':
			hex2mem(&remcomInBuffer[1],
				(char *)gdb_regs, NUMREGBYTES, 0);
			if (!usethread || usethread == current) {
				gdb_regs_to_regs(gdb_regs, regs);
				strcpy(remcomOutBuffer, "OK");
			} else {
				strcpy(remcomOutBuffer, "E00");
			}
			break;

		case 'P':{	/* set the value of a single CPU register -
				   return OK */
				/*
				 * For some reason, gdb wants to talk about psudo
				 * registers (greater than 15).	 These may have
				 * meaning for ptrace, but for us it is safe to
				 * ignor them.	We do this by dumping them into
				 * _GS which we also ignor, but do have memory for.
				 */
				int regno;

				ptr = &remcomInBuffer[1];
				regs_to_gdb_regs(gdb_regs, regs);
				if ((!usethread || usethread == current) &&
				    hexToInt(&ptr, &regno) &&
				    *ptr++ == '=' && (regno >= 0)) {
					regno =
					    (regno >= NUMREGS ? _GS : regno);
					hex2mem(ptr, (char *) &gdb_regs[regno],
						4, 0);
					gdb_regs_to_regs(gdb_regs, regs);
					strcpy(remcomOutBuffer, "OK");
					break;
				}
				strcpy(remcomOutBuffer, "E01");
				break;
			}

			/* mAA..AA,LLLL	 Read LLLL bytes at address AA..AA */
		case 'm':
			/* TRY TO READ %x,%x.  IF SUCCEED, SET PTR = 0 */
			ptr = &remcomInBuffer[1];
			if (hexToInt(&ptr, &addr) &&
			    (*(ptr++) == ',') && (hexToInt(&ptr, &length))) {
				ptr = 0;
				/*
				 * hex doubles the byte count
				 */
				if (length > (BUFMAX / 2))
					length = BUFMAX / 2;
				mem2hex((char *) addr,
					remcomOutBuffer, length, 1);
				if (mem_err) {
					strcpy(remcomOutBuffer, "E03");
					debug_error("memory fault\n", NULL);
				}
			}

			if (ptr) {
				strcpy(remcomOutBuffer, "E01");
				debug_error
				    ("malformed read memory command: %s\n",
				     remcomInBuffer);
			}
			break;

			/* MAA..AA,LLLL:
			   Write LLLL bytes at address AA.AA return OK */
		case 'M':
			/* TRY TO READ '%x,%x:'.  IF SUCCEED, SET PTR = 0 */
			ptr = &remcomInBuffer[1];
			if (hexToInt(&ptr, &addr) &&
			    (*(ptr++) == ',') &&
			    (hexToInt(&ptr, &length)) && (*(ptr++) == ':')) {
				hex2mem(ptr, (char *) addr, length, 1);

				if (mem_err) {
					strcpy(remcomOutBuffer, "E03");
					debug_error("memory fault\n", NULL);
				} else {
					strcpy(remcomOutBuffer, "OK");
				}

				ptr = 0;
			}
			if (ptr) {
				strcpy(remcomOutBuffer, "E02");
				debug_error
				    ("malformed write memory command: %s\n",
				     remcomInBuffer);
			}
			break;
		case 'S':
			remcomInBuffer[0] = 's';
		case 'C':
			/* Csig;AA..AA where ;AA..AA is optional
			 * continue with signal
			 * Since signals are meaning less to us, delete that
			 * part and then fall into the 'c' code.
			 */
			ptr = &remcomInBuffer[1];
			length = 2;
			while (*ptr && *ptr != ';') {
				length++;
				ptr++;
			}
			if (*ptr) {
				do {
					ptr++;
					*(ptr - length++) = *ptr;
				} while (*ptr);
			} else {
				remcomInBuffer[1] = 0;
			}

			/* cAA..AA  Continue at address AA..AA(optional) */
			/* sAA..AA  Step one instruction from AA..AA(optional)*/
			/* D	    detach, reply OK and then continue */
		case 'c':
		case 's':
		case 'D':

			/* try to read optional parameter,
			   pc unchanged if no parm */
			ptr = &remcomInBuffer[1];
			if (hexToInt(&ptr, &addr)) {
				if (kgdb_info.print_debug_info)
					printk("Changing EIP to 0x%x\n", addr);

				regs->eip = addr;
			}

			newPC = regs->eip;

			/* clear the trace bit */
			regs->eflags &= 0xfffffeff;

			/* set the trace bit if we're stepping */
			if (remcomInBuffer[0] == 's')
				regs->eflags |= 0x100;

			/* detach is a friendly version of continue. Note that
			   debugging is still enabled (e.g hit control C)
			 */
			if (remcomInBuffer[0] == 'D') {
				strcpy(remcomOutBuffer, "OK");
				putpacket(remcomOutBuffer);
			}

			if (kgdb_info.print_debug_info) {
				printk("Resuming execution\n");
				print_regs(regs);
			}
			asm volatile ("movl %%db6, %0\n":"=r" (dr6)
				      :);
			if (!(dr6 & 0x4000)) {
				for (breakno = 0; breakno < 4; ++breakno) {
					if (dr6 & (1 << breakno) &&
					    (breakinfo[breakno].type == 0)) {
						/* Set restore flag */
						regs->eflags |= 0x10000;
						break;
					}
				}
			}
			correct_hw_break();
			asm volatile ("movl %0, %%db6\n"::"r" (0));
			goto exit_kgdb;

			/* kill the program */
		case 'k':	/* do nothing */
			break;

			/* query */
		case 'q':
			nothreads = 0;
			switch (remcomInBuffer[1]) {
			case 'f':
				threadid = 1;
				thread_list = 2;
				thread_list_start = (usethread ? : current);
			case 's':
				if (!cmp_str(&remcomInBuffer[2],
					     "ThreadInfo", 10))
					break;

				remcomOutBuffer[nothreads++] = 'm';
				for (; threadid < PID_MAX + NR_CPUS;
				     threadid++) {
					thread = getthread(threadid);
					if (thread) {
						nothreads += int_to_hex_v(
							&remcomOutBuffer[
								nothreads],
							threadid);
						if (thread_min > threadid)
							thread_min = threadid;
						remcomOutBuffer[
							nothreads] = ',';
						nothreads++;
						if (nothreads > BUFMAX - 10)
							break;
					}
				}
				if (remcomOutBuffer[nothreads - 1] == 'm')
					remcomOutBuffer[nothreads - 1] = 'l';
				else
					nothreads--;
				remcomOutBuffer[nothreads] = 0;
				break;

			case 'C':
				/* Current thread id */
				remcomOutBuffer[0] = 'Q';
				remcomOutBuffer[1] = 'C';
				threadid = current->pid;
				if (!threadid) {
					/*
					 * idle thread
					 */
					for (threadid = PID_MAX;
					     threadid < PID_MAX + NR_CPUS;
					     threadid++) {
						if (current ==
						    kgdb_get_idle(threadid-PID_MAX))
							break;
					}
				}
				int_to_threadref(&thref, threadid);
				pack_threadid(remcomOutBuffer + 2, &thref);
				remcomOutBuffer[18] = '\0';
				break;

			case 'E':
				/* Print exception info */
				printexceptioninfo(exceptionVector,
						   err_code, remcomOutBuffer);
				break;
			case 'T':{
				char * nptr;
				/* Thread extra info */
				if (!cmp_str(&remcomInBuffer[2],
					    "hreadExtraInfo,", 15))
					break;
				ptr = &remcomInBuffer[17];
				hexToInt(&ptr, &threadid);
				thread = getthread(threadid);
				nptr = &thread->comm[0];
				length = 0;
				ptr = &remcomOutBuffer[0];
				do {
					length++;
					ptr = pack_hex_byte(ptr, *nptr++);
				} while (*nptr && length < 16);
				/*
				 * would like that 16 to be the size of
				 * task_struct.comm but don't know the
				 * syntax..
				 */
				*ptr = 0;
			}
			}
			break;

			/* task related */
		case 'H':
			switch (remcomInBuffer[1]) {
			case 'g':
				ptr = &remcomInBuffer[2];
				hexToInt(&ptr, &threadid);
				thread = getthread(threadid);
				if (!thread) {
					remcomOutBuffer[0] = 'E';
					remcomOutBuffer[1] = '\0';
					break;
				}
				/*
				 * Just in case I forget what this is all about,
				 * the "thread info" command to gdb causes it
				 * to ask for a thread list.  It then switches
				 * to each thread and asks for the registers.
				 * For this (and only this) usage, we want to
				 * fudge the registers of tasks not on the run
				 * list (i.e. waiting) to show the routine that
				 * called schedule. Also, gdb, is a minimalist
				 * in that if the current thread is the last
				 * it will not re-read the info when done.
				 * This means that in this case we must show
				 * the real registers. So here is how we do it:
				 * Each entry we keep track of the min
				 * thread in the list (the last that gdb will)
				 * get info for.  We also keep track of the
				 * starting thread.
				 * "thread_list" is cleared when switching back
				 * to the min thread if it is was current, or
				 * if it was not current, thread_list is set
				 * to 1.  When the switch to current comes,
				 * if thread_list is 1, clear it, else do
				 * nothing.
				 */
				usethread = thread;
				if ((thread_list == 1) &&
						(thread == thread_list_start))
					thread_list = 0;
				if (thread_list && (threadid == thread_min)) {
					if (thread == thread_list_start)
						thread_list = 0;
					else
						thread_list = 1;
				}
				/* follow through */
			case 'c':
				remcomOutBuffer[0] = 'O';
				remcomOutBuffer[1] = 'K';
				remcomOutBuffer[2] = '\0';
				break;
			}
			break;

			/* Query thread status */
		case 'T':
			ptr = &remcomInBuffer[1];
			hexToInt(&ptr, &threadid);
			thread = getthread(threadid);
			if (thread) {
				remcomOutBuffer[0] = 'O';
				remcomOutBuffer[1] = 'K';
				remcomOutBuffer[2] = '\0';
				if (thread_min > threadid)
					thread_min = threadid;
			} else {
				remcomOutBuffer[0] = 'E';
				remcomOutBuffer[1] = '\0';
			}
			break;

		case 'Y': /* set up a hardware breakpoint */
			ptr = &remcomInBuffer[1];
			hexToInt(&ptr, &breakno);
			ptr++;
			hexToInt(&ptr, &breaktype);
			ptr++;
			hexToInt(&ptr, &length);
			ptr++;
			hexToInt(&ptr, &addr);
			if (set_hw_break(breakno & 0x3,
					 breaktype & 0x3,
					 length & 0x3, addr) == 0) {
				strcpy(remcomOutBuffer, "OK");
			} else {
				strcpy(remcomOutBuffer, "ERROR");
			}
			break;

			/* Remove hardware breakpoint */
		case 'y':
			ptr = &remcomInBuffer[1];
			hexToInt(&ptr, &breakno);
			if (remove_hw_break(breakno & 0x3) == 0) {
				strcpy(remcomOutBuffer, "OK");
			} else {
				strcpy(remcomOutBuffer, "ERROR");
			}
			break;

		case 'r':	/* reboot */
			strcpy(remcomOutBuffer, "OK");
			putpacket(remcomOutBuffer);
			/*to_gdb("Rebooting\n"); */
			/* triplefault	 no return from here */
			__asm__ __volatile__("lidt %0"::"m"(no_idt[0]));
			KGDB_BREAKPOINT();
		}
		putpacket(remcomOutBuffer);	/* reply to the request */
	}
	/*
	 *  reached by goto only.
	 */
exit_kgdb:
	/*
	 * Here is where we set up to trap a gdb function call.	 fn_call_lookaside[trap_cpu].esp
	 * will be changed if we are trying to do this.	 We handle both
	 * adding and subtracting, thus allowing gdb to put grung on
	 * the stack which it removes later.
	 */
	if (fn_call_lookaside[trap_cpu].esp != OLD_esp) {
		int *ptr = END_OF_LOOKASIDE;
		if (fn_call_lookaside[trap_cpu].esp < OLD_esp)
			ptr -= (OLD_esp - fn_call_lookaside[trap_cpu].esp) / sizeof (int);
		*--ptr = regs->eflags;
		*--ptr = regs->xcs;
		*--ptr = regs->eip;
		*--ptr = regs->ecx;
		*--ptr = regs->ebx;
		*--ptr = regs->eax;
		regs->ecx = fn_call_lookaside[trap_cpu].esp - (sizeof (int) * 6);
		regs->ebx = (unsigned int) END_OF_LOOKASIDE;
		if (fn_call_lookaside[trap_cpu].esp < OLD_esp) {
			regs->eip = (unsigned int) fn_call_stub;
		} else {
			regs->eip = (unsigned int) fn_rtn_stub;
			regs->eax = fn_call_lookaside[trap_cpu].esp;
		}
		regs->eflags &= ~(IF_BIT | TF_BIT);
	}

	/*
	 * Release gdb wait locks
	 * Sanity check time.  Must have at least one cpu to run.  Also single
	 * step must not be done if the current cpu is on hold.
	 */
	if (spinlock_count == 1) {
		int ss_hold = (regs->eflags & 0x100) &&
						kgdb_info.hold_on_sstep;
		int cpu_avail = 0;
		int i;

		for_each_online_cpu(i) {
			if (!kgdb_info.cpus_waiting[i].hold)
				cpu_avail = 1;
		}
		/*
		 * Early in the bring up there will be NO cpus on line...
		 */
		if (!cpu_avail && !cpus_empty(cpu_online_map)) {
			to_gdb("No cpus unblocked, see 'kgdb_info.hold_cpu'\n");
			goto once_again;
		}
		if (kgdb_info.cpus_waiting[smp_processor_id()].hold &&
				(regs->eflags & 0x100)) {
			to_gdb
			    ("Current cpu must be unblocked to single step\n");
			goto once_again;
		}
		if (!ss_hold) {
			for_each_online_cpu(i) {
				if (!kgdb_info.cpus_waiting[i].hold)
					spin_unlock(&waitlocks[i]);
			}
		} else {
			spin_unlock(&waitlocks[smp_processor_id()]);
		}
		/* Release kgdb spinlock */
		KGDB_SPIN_UNLOCK(&kgdb_spinlock);
		/*
		 * If this cpu is on hold, this is where we
		 * do it.  Note, the NMI will pull us out of here,
		 * but will return as the above lock is not held.
		 * We will stay here till another cpu releases the lock for us.
		 */
		spin_unlock_wait(waitlocks + smp_processor_id());
		local_irq_restore(flags);
		return;
	}
	/* Release kgdb spinlock */
	KGDB_SPIN_UNLOCK(&kgdb_spinlock);
	local_irq_restore(flags);
	return;
}

static void kgdb_console_write(struct console *co, const char *s,
				unsigned count)
{

	if (kgdb_info.vector == -1) {
		/*
		 * We have not yet talked to gdb.  What to do...
		 * lets break, on continue we can do the write.
		 * But first tell him whats up. Uh, well no can do,
		 * as this IS the console.  Oh well...
		 * We do need to wait or the messages will be lost.
		 * Other option would be to tell the above code to
		 * ignore this breakpoint and do an auto return,
		 * but that might confuse gdb.	Also this happens
		 * early enough in boot up that we don't have the traps
		 * set up yet, so...
		 */
		kgdb_breakpoint();
	}
	kgdb_gdb_message(s, count);
}

/*
 * ------------------------------------------------------------
 * Serial KGDB driver
 * ------------------------------------------------------------
 */

static struct console kgdbcons = {
	.name = "kgdb",
	.write= kgdb_console_write,
	.flags = CON_PRINTBUFFER | CON_ENABLED,
	.index = -1,
};

/*
 * The trick here is that this file gets linked before printk.o
 * That means we get to peer at the console info in the command
 * line before it does.	 If we are up, we register, otherwise,
 * do nothing.	By returning 0, we allow printk to look also.
 */
static int kgdb_console_enabled;

int __init kgdb_console_init(char *str)
{
	if ((strncmp(str, "kgdb", 4) == 0) || (strncmp(str, "gdb", 3) == 0)) {
		register_console(&kgdbcons);
		kgdb_console_enabled = 1;
	}
	return 0;		/* let others look at the string */
}
__setup("console=", kgdb_console_init);

typedef void gdb_debug_hook(int exceptionVector,
			int signo, int err_code, struct pt_regs *linux_regs);
gdb_debug_hook *linux_debug_hook = &kgdb_handle_exception;

/*
 * Provide the command line "gdb" initial break
 */
int __init kgdb_initial_break(char * str)
{
	kgdb_enabled = 1;
	if (*str == '\0'){
		kgdb_breakpoint();
		return 1;
	}
	return 0;
}
__setup("gdb",kgdb_initial_break);
