/*
 * Serial interface GDB stub
 *
 * Written (hacked together) by David Grothe (dave@gcom.com)
 * Modified to allow invokation early in boot see also
 * kgdb.h for instructions by George Anzinger(george@mvista.com)
 *
 */

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <linux/serial_reg.h>
#include <linux/config.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/init.h>
#include <linux/highmem.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/bitops.h>
#include <asm/system.h>
#include <asm/kgdb_local.h>

#define PRNT_off
#ifdef PRNT
#define dbprintk(s) printk s
#else
#define dbprintk(s)
#endif
#define TEST_INTERRUPT_off
#ifdef TEST_INTERRUPT
#define intprintk(s) printk s
#else
#define intprintk(s)
#endif

static char gdb_buf[512];
static int gdb_buf_in_inx;
static atomic_t gdb_buf_in_cnt;
static int gdb_buf_out_inx;

struct async_struct *gdb_async_info;
static int gdb_async_irq;

static void program_uart(struct async_struct *info);
static void write_char(struct async_struct *info, int chr);

/*
 * Get a byte from the hardware data buffer and return it
 */
static int read_data_bfr(struct async_struct *info)
{
	char it = inb_p(info->port + UART_LSR);

	if (it & UART_LSR_DR)
		return (inb_p(info->port + UART_RX));

	/*
	 * If we have a framing error assume somebody messed with
	 * our uart.  Reprogram it and send '-' both ways...
	 */
	if (it & 0xc) {
		program_uart(info);
		write_char(info, '-');
		return ('-');
	}
	return (-1);

}				/* read_data_bfr */

/*
 * Get a char if available, return -1 if nothing available.
 * Empty the receive buffer first, then look at the interface hardware.

 * Locking here is a bit of a problem.	We MUST not lock out communication
 * if we are trying to talk to gdb about a kgdb entry.	ON the other hand
 * we can loose chars in the console pass thru if we don't lock.  It is also
 * possible that we could hold the lock or be waiting for it when kgdb
 * NEEDS to talk.  Since kgdb locks down the world, it does not need locks.
 * We do, of course have possible issues with interrupting a uart operation,
 * but we will just depend on the uart status to help keep that straight.

 */
static DEFINE_SPINLOCK(uart_interrupt_lock);
extern spinlock_t kgdb_spinlock;

static int read_char(struct async_struct *info)
{
	int chr;
	unsigned long flags;
	local_irq_save(flags);

	if (!spin_is_locked(&kgdb_spinlock))
		spin_lock(&uart_interrupt_lock);

	if (atomic_read(&gdb_buf_in_cnt) != 0) {
		/* intr routine has q'd chars */
		chr = gdb_buf[gdb_buf_out_inx++];
		gdb_buf_out_inx &= (ARRAY_SIZE(gdb_buf) - 1);
		atomic_dec(&gdb_buf_in_cnt);
	} else {
		chr = read_data_bfr(info);
	}

	if (!spin_is_locked(&kgdb_spinlock))
		spin_unlock(&uart_interrupt_lock);

	local_irq_restore(flags);
	return (chr);
}

/*
 * Wait until the interface can accept a char, then write it.
 */
static void write_char(struct async_struct *info, int chr)
{
	while (!(inb_p(info->port + UART_LSR) & UART_LSR_THRE))
		;
	outb_p(chr, info->port + UART_TX);
}

/*
 * Mostly we don't need a spinlock, but since the console goes
 * thru here with interrutps on, well, we need to catch those
 * chars.
 */
/*
 * This is the receiver interrupt routine for the GDB stub.
 * It will receive a limited number of characters of input
 * from the gdb  host machine and save them up in a buffer.
 *
 * When the gdb stub routine getDebugChar() is called it
 * draws characters out of the buffer until it is empty and
 * then reads directly from the serial port.
 *
 * We do not attempt to write chars from the interrupt routine
 * since the stubs do all of that via putDebugChar() which
 * writes one byte after waiting for the interface to become
 * ready.
 *
 * The debug stubs like to run with interrupts disabled since,
 * after all, they run as a consequence of a breakpoint in
 * the kernel.
 *
 * Perhaps someone who knows more about the tty driver than I
 * care to learn can make this work for any low level serial
 * driver.
 */
static irqreturn_t gdb_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct async_struct *info;
	unsigned long flags;

	info = gdb_async_info;
	if (!info || !info->tty || irq != gdb_async_irq)
		return IRQ_NONE;

	local_irq_save(flags);
	spin_lock(&uart_interrupt_lock);
	do {
		int chr = read_data_bfr(info);
		intprintk(("Debug char on int: %x hex\n", chr));
		if (chr < 0)
			continue;

		if (chr == 3) {	/* Ctrl-C means remote interrupt */
			KGDB_BREAKPOINT();
			continue;
		}

		if (atomic_read(&gdb_buf_in_cnt) >= ARRAY_SIZE(gdb_buf)) {
			/* buffer overflow tosses early char */
			read_char(info);
		}
		gdb_buf[gdb_buf_in_inx++] = chr;
		gdb_buf_in_inx &= (ARRAY_SIZE(gdb_buf) - 1);
	} while (inb_p(info->port + UART_IIR) & UART_IIR_RDI);
	spin_unlock(&uart_interrupt_lock);
	local_irq_restore(flags);
	return IRQ_HANDLED;
}

/* These structure are filled in with values defined in asm/kgdb_local.h */
static struct serial_state state = SB_STATE;
static struct async_struct local_info = SB_INFO;
static int ok_to_enable_ints;
static void kgdb_enable_ints_now(void);

/*
 * Hook an IRQ for KGDB.
 *
 * This routine is called from putDebugChar, below.
 */
static int ints_disabled = 1;

int gdb_hook_interrupt(struct async_struct *info, int verb)
{
	struct serial_state *state = info->state;
	unsigned long flags;
	int port;
	int scratch, scratch2;

	/* The above fails if memory managment is not set up yet.
	 * Rather than fail the set up, just keep track of the fact
	 * and pick up the interrupt thing later.
	 */
	gdb_async_info = info;
	port = gdb_async_info->port;
	gdb_async_irq = state->irq;
	if (verb) {
		printk("kgdb: port =%x, IRQ=%d, divisor =%d\n",
			port, gdb_async_irq,
			gdb_async_info->state->custom_divisor);
	}
	local_irq_save(flags);

	/* Existence test.  Should not need all this, but just in case.... */
	scratch = inb_p(port + UART_IER);
	outb_p(0, port + UART_IER);
	outb_p(0x080, 0xff);
	scratch2 = inb_p(port + UART_IER);
	outb_p(scratch, port + UART_IER);
	if (scratch2) {
		printk("gdb_hook_interrupt: Could not clear IER, "
				"not a UART!\n");
		local_irq_restore(flags);
		return 1;	/* We failed; there's nothing here */
	}
	scratch2 = inb_p(port + UART_LCR);
	outb_p(0xBF, port + UART_LCR);	/* set up for StarTech test */
	outb_p(0, port + UART_EFR);	/* EFR is the same as FCR */
	outb_p(0, port + UART_LCR);
	outb_p(UART_FCR_ENABLE_FIFO, port + UART_FCR);
	scratch = inb_p(port + UART_IIR) >> 6;
	if (scratch == 1) {
		printk("gdb_hook_interrupt: Undefined UART type!"
		       "  Not a UART! \n");
		local_irq_restore(flags);
		return 1;
	} else {
		dbprintk(("gdb_hook_interrupt: UART type "
			  "is %d where 0=16450, 2=16550 3=16550A\n", scratch));
	}
	scratch = inb_p(port + UART_MCR);
	outb_p(UART_MCR_LOOP | scratch, port + UART_MCR);
	outb_p(UART_MCR_LOOP | 0x0A, port + UART_MCR);
	scratch2 = inb_p(port + UART_MSR) & 0xF0;
	outb_p(scratch, port + UART_MCR);
	if (scratch2 != 0x90) {
		printk("gdb_hook_interrupt: "
		       "Loop back test failed! Not a UART!\n");
		local_irq_restore(flags);
		return scratch2 + 1000;	/* force 0 to fail */
	}
	program_uart(info);
	local_irq_restore(flags);
	return 0;
}

static void program_uart(struct async_struct *info)
{
	int port = info->port;

	inb_p(port + UART_RX);
	outb_p(0, port + UART_IER);

	inb_p(port + UART_RX);	/* serial driver comments say */
	inb_p(port + UART_IIR);	/* this clears the interrupt regs */
	inb_p(port + UART_MSR);
	outb_p(UART_LCR_WLEN8 | UART_LCR_DLAB, port + UART_LCR);
	outb_p(info->state->custom_divisor & 0xff, port + UART_DLL); /* LS */
	outb_p(info->state->custom_divisor >> 8, port + UART_DLM);   /* MS  */
	outb_p(info->MCR, port + UART_MCR);

	outb_p(UART_FCR_ENABLE_FIFO|UART_FCR_TRIGGER_1|UART_FCR_CLEAR_XMIT|
		UART_FCR_CLEAR_RCVR, port + UART_FCR);	/* set fcr */
	outb_p(UART_LCR_WLEN8, port + UART_LCR);	/* reset DLAB */
	outb_p(UART_FCR_ENABLE_FIFO|UART_FCR_TRIGGER_1, port + UART_FCR);
	if (!ints_disabled) {
		intprintk(("KGDB: Sending %d to port %x offset %d\n",
			   gdb_async_info->IER,
			   (int) gdb_async_info->port, UART_IER));
		outb_p(gdb_async_info->IER, gdb_async_info->port + UART_IER);
	}
	return;
}

/*
 * getDebugChar
 *
 * This is a GDB stub routine.	It waits for a character from the
 * serial interface and then returns it.  If there is no serial
 * interface connection then it returns a bogus value which will
 * almost certainly cause the system to hang.  In the
 */
int kgdb_in_isr;
int kgdb_in_lsr;

/* Caller takes needed protections */

int getDebugChar(void)
{
	volatile int chr, dum, time, end_time;

	dbprintk(("getDebugChar(port %x): ", gdb_async_info->port));

	if (gdb_async_info == NULL)
		gdb_hook_interrupt(&local_info, 0);

	/*
	 * This trick says if we wait a very long time and get
	 * no char, return the -1 and let the upper level deal
	 * with it.
	 */
	rdtsc(dum, time);
	end_time = time + 2;
	while (((chr = read_char(gdb_async_info)) == -1) &&
			(end_time - time) > 0)
		rdtsc(dum, time);
	/*
	 * This covers our butts if some other code messes with
	 * our uart, hay, it happens :o)
	 */
	if (chr == -1)
		program_uart(gdb_async_info);

	dbprintk(("%c\n", chr > ' ' && chr < 0x7F ? chr : ' '));
	return chr;
}

static int count = 3;

static int __init kgdb_enable_ints(void)
{
	if (gdb_async_info == NULL)
		gdb_hook_interrupt(&local_info, 1);
	ok_to_enable_ints = 1;
	kgdb_enable_ints_now();
	return 0;
}

#ifdef CONFIG_SERIAL_8250
void shutdown_for_kgdb(struct async_struct *gdb_async_info);
#else
static inline void shutdown_for_kgdb(struct async_struct *gdb_async_info) {}
#endif

#ifdef CONFIG_DISCONTIGMEM
static inline int kgdb_mem_init_done(void)
{
	return totalram_pages != 0;
}
#else
static inline int kgdb_mem_init_done(void)
{
	return max_mapnr != 0;
}
#endif

static void kgdb_enable_ints_now(void)
{
	static DEFINE_SPINLOCK(one_at_atime);

	if (!spin_trylock(&one_at_atime))
		return;
	if (!ints_disabled)
		goto exit;
	if (kgdb_mem_init_done() && ints_disabled) {
		/* don't try till mem init */
		/*
		 * The ifdef here allows the system to be configured
		 * without the serial driver.
		 * Don't make it a module, however, it will steal the port
		 */
		shutdown_for_kgdb(gdb_async_info);
		ints_disabled = request_irq(gdb_async_info->state->irq,
				gdb_interrupt,
				(gdb_async_info->flags & ASYNC_SHARE_IRQ) ?
					SA_SHIRQ : SA_INTERRUPT,
				"KGDB-stub", NULL);
		intprintk(("KGDB: request_irq returned %d\n", ints_disabled));
	}
	if (!ints_disabled) {
		intprintk(("KGDB: Sending %d to port %x offset %d\n",
			   gdb_async_info->IER,
			   (int) gdb_async_info->port, UART_IER));
		outb_p(gdb_async_info->IER, gdb_async_info->port + UART_IER);
	}
exit:
	spin_unlock(&one_at_atime);
}

/*
 * putDebugChar
 *
 * This is a GDB stub routine.	It waits until the interface is ready
 * to transmit a char and then sends it.  If there is no serial
 * interface connection then it simply returns to its caller, having
 * pretended to send the char.	Caller takes needed protections.
 */
void putDebugChar(int chr)
{
	dbprintk(("putDebugChar(port %x): chr=%02x '%c', ints_on=%d\n",
		  gdb_async_info->port,
		  chr,
		  chr > ' ' && chr < 0x7F ? chr : ' ', ints_disabled ? 0 : 1));

	if (gdb_async_info == NULL)
		gdb_hook_interrupt(&local_info, 0);

	write_char(gdb_async_info, chr);	/* this routine will wait */
	count = (chr == '#') ? 0 : count + 1;
	if (count == 2) {	/* try to enable after */
		if (ints_disabled & ok_to_enable_ints)
			kgdb_enable_ints_now();	/* try to enable after */

		/*
		 * We do this a lot because, well we really want to get these
		 * interrupts.	The serial driver will clear these bits when it
		 * initializes the chip.  Every thing else it does is ok,
		 * but this.
		 */
		if (!ints_disabled)
			outb_p(gdb_async_info->IER,
				gdb_async_info->port + UART_IER);
	}
}

module_init(kgdb_enable_ints);
